import logging
import os
import subprocess
import unittest
import warnings

import pytest
from s_code_project_updater.commands import UpdateException
from schema import create_session
from schema.model import Project
from schema.pstmodel import Session
from sqlalchemy import exc as sa_exc
from support.logging import get_console_logger

from .test_projects import get_author_pst_ids, ScodeTestProject, \
    ScienceTestProject, AlmaTestProject

_LOG = get_console_logger("scode_project_updater_tests", logging.DEBUG)
_UPDATE_COMMAND = 'update_sproj'
PROFILE = 'local'


class UpdaterTestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        os.environ['CAPO_PROFILE'] = PROFILE
        cls.return_values = build_updater_return_values()

    def setUp(self) -> None:
        self.initialize_test_data()

    def test_dry_run_does_not_update(self):
        fake_project = ScodeTestProject().project
        project_code = fake_project.project_code
        session = create_session('SDM')
        return_code = None
        try:
            new_title = 'this is the new title'
            self.assertNotEqual(fake_project.title, new_title)
            args = [
                '-C', project_code,
                '-P', PROFILE,
                '-T', new_title,
                '--dry'
                ]
            try:
                return_code = CommandLineUpdaterLauncher(args).run()
            except Exception as exc:
                text = self.return_values[return_code] if return_code else ''
                pytest.fail(f'{exc} {text}')

            if not return_code:
                updated = self.get_project_from_db(session, project_code)
                # nothing should have been updated
                self.assertEqual(fake_project.title, updated.title)
                self.assertEqual(fake_project.abstract, updated.abstract)
                self.assertEqual(len(fake_project.authors),
                                     len(updated.authors))

        except Exception as exc:
            pytest.fail(f'{project_code}: {exc}')
        finally:
            session.close()

    def test_project_code_only_fetches(self):
        fake_project = ScodeTestProject().project
        project_code = fake_project.project_code
        args = [
            '-C', project_code,
            '-P', PROFILE,
        ]
        return_code = None
        try:
            return_code = CommandLineUpdaterLauncher(args).run()
            if not return_code:
                session = create_session('SDM')
                try:
                    updated = self.get_project_from_db(session, project_code)
                    self.assertEqual(fake_project.title, updated.title)
                    self.assertEqual(fake_project.abstract, updated.abstract)
                    self.assertEqual(len(fake_project.authors),
                                     len(updated.authors))
                    count = 0
                    for orig_author in fake_project.authors:
                        for author in updated.authors:
                            if author.username == orig_author.username:
                                count += 1
                                break
                    self.assertEqual(len(fake_project.authors), count)
                finally:
                    session.close()
        except Exception as exc:
            text = self.return_values[return_code] if return_code else ''
            pytest.fail(f'{exc} {text}')

    def test_updates_abstract_only(self):
        fake_project = ScodeTestProject().project
        project_code = fake_project.project_code
        session = create_session('SDM')
        new_abstract = "Well, here's another nice mess you've gotten us into, Ollie"
        self.assertNotEqual(fake_project.abstract, new_abstract)
        args = [
            '-C', project_code,
            '-P', PROFILE,
            '-A', new_abstract,
        ]
        return_code = None
        try:
            try:
                return_code = CommandLineUpdaterLauncher(args).run()
            except subprocess.TimeoutExpired as exp:
                raise UpdateException(exp)
            except Exception as exc:
                text = self.return_values[return_code] if return_code else ''
                pytest.fail(f'{exc} {text}')

            if not return_code:
                updated = self.get_project_from_db(session, project_code)
                # only abstract should have been updated;
                # all else should be same
                self.assertEqual(fake_project.title, updated.title)
                self.assertEqual(new_abstract, updated.abstract)
                self.assertEqual(len(fake_project.authors),
                                 len(updated.authors))
            else:
                raise UpdateException()

        except Exception as exc:
            pytest.fail(f'{project_code}: {exc}')
        finally:
            session.close()

    def test_updates_abstract_and_title(self):
        fake_project = ScodeTestProject().project
        project_code = fake_project.project_code
        session = create_session('SDM')
        new_abstract = "I think you ought to know I'm feeling very depressed"
        new_title = 'A Survey of the Mattresses of Sqornshellous Zeta'
        self.assertNotEqual(fake_project.abstract, new_abstract)
        self.assertNotEqual(fake_project.title, new_title)
        args = [
            '-C', project_code,
            '-P', PROFILE,
            '-A', new_abstract,
            '-T', new_title,
        ]
        return_code = None
        try:
            try:
                return_code = CommandLineUpdaterLauncher(args).run()
            except subprocess.TimeoutExpired as exp:
                raise UpdateException(exp)
            except Exception as exc:
                text = self.return_values[return_code] if return_code else ''
                pytest.fail(f'{exc} {text}')

            if not return_code:
                updated = self.get_project_from_db(session, project_code)
                # abstract and title should have been updated;
                # all else should be same
                self.assertEqual(new_title, updated.title)
                self.assertEqual(new_abstract, updated.abstract)
                self.assertEqual(len(fake_project.authors),
                                 len(updated.authors))
            else:
                raise UpdateException()

        except Exception as exc:
            pytest.fail(f'{project_code}: {exc}')
        finally:
            session.close()

    def test_adds_new_abstract_deletes_author(self):
        fake_project = ScodeTestProject().project
        new_project = fake_project
        project_code = fake_project.project_code
        new_abstract = "First there is a mountain, then there is no " \
                        "mountain, then there is"
        self.assertNotEqual(new_abstract, fake_project.abstract)
        new_project.abstract = new_abstract
        original_authors = fake_project.authors.copy()
        new_authors = original_authors[:3]
        self.assertEqual(len(original_authors) - 1, len(new_authors))
        new_project.authors = new_authors
        args = [
            '-C', project_code,
            '-P', PROFILE,
            '-A', new_abstract,
            '-I',
        ]
        for id in get_author_pst_ids(new_project):
            args.append(str(id))

        return_code = None
        session = create_session('SDM')
        try:
            try:
                return_code = CommandLineUpdaterLauncher(args).run()
            except subprocess.TimeoutExpired as exp:
                raise UpdateException(exp)
            except Exception as exc:
                text = self.return_values[return_code] if return_code else ''
                pytest.fail(f'{exc} {text}')

            if not return_code:
                updated = self.get_project_from_db(session, project_code)
                # last author should have been removed and the abstract changed;
                # title should remain same
                self.assertEqual(new_abstract, updated.abstract)
                self.assertEqual(fake_project.title, updated.title)
                self.assertEqual(len(new_authors), len(updated.authors))
                self.assertFalse(original_authors[3] in updated.authors)
                count = 0
                for orig_author in original_authors[:3]:
                    for new_author in updated.authors:
                        if new_author.username == orig_author.username:
                            count += 1
                            break
                self.assertEqual(len(new_authors), count)
            else:
                raise UpdateException()

        except Exception as exc:
            pytest.fail(f'{project_code}: {exc}')

        finally:
            session.close()


    ### UTILITIES ###

    def initialize_test_data(self):
        session = create_session('SDM')
        try:
            with warnings.catch_warnings():
                # Suppress SQLAlchemy warnings
                warnings.simplefilter("ignore", category=sa_exc.SAWarning)

                fake_projects = [ScodeTestProject().project,
                                 ScienceTestProject().project,
                                 AlmaTestProject().project]
                num_commits = 0
                try:
                    for fake_project in fake_projects:
                        project_code = fake_project.project_code
                        existing = session.query(Project) \
                            .filter(Project.project_code ==
                                    project_code) \
                            .first()
                        if existing is not None:
                            session.delete(existing)
                            session.commit()
                        session.add(fake_project)
                        session.commit()
                        num_commits += 1

                    self.assertEqual(len(fake_projects), num_commits)
                except Exception as exc:
                    pytest.fail(f'{exc}')
        finally:
            session.close()

    def get_project_from_db(self, session: Session, project_code: str):
        with warnings.catch_warnings():
            # Suppress SQLAlchemy warnings
            warnings.simplefilter("ignore", category=sa_exc.SAWarning)

            return session.query(Project) \
                .filter(Project.project_code == project_code) \
                .first()

class CommandLineUpdaterLauncher:

    def __init__(self, args: list):
        self.args = [_UPDATE_COMMAND]
        for arg in args:
            self.args.append(str(arg))
        _LOG.info(f'{self.args}')

    def run(self):
        ''' launch updater from command line
            @:returns directory listing
        '''

        args = self.args
        try:
            proc = subprocess.run(args,
                                       stdout=subprocess.PIPE,
                                       stderr=subprocess.STDOUT,
                                       timeout=60,
                                       check=False,
                                       bufsize=1,
                                       universal_newlines=True)
            return proc.returncode
        except Exception as exc:
            _LOG.error(f'{exc}')
            if not isinstance(exc, subprocess.TimeoutExpired):
                return exc.returncode
            else:
                raise

def build_updater_return_values():
    return {
        1: 'error with capo configuration',
        2: 'error with input parameters',
        3: 'project not found',
        4: 'investigator not found',
        5: 'update failed',
    }

if __name__ == '__main__':
    unittest.main()
