import json
import os

import datetime
import pika
from pycapo import CapoConfig

EVENTS_EXCHANGE = 'archive.events'

""""
A simple API for sending an archive event from Python.
"""

# Local variables
_connection, _channel = None, None

def broadcast(message, **kwargs):
    """
    Broadcasts a message to the archive event system.

    Keyword arguments for the message:

    :param: application: the application sending the message (defaults to 'python')
    :param: user: the user sending the event (defaults to the currently logged in user)
    :param: request: the request this event is in reference to, if applicable
    :param: message: the message itself (will override the formal parameter)
    :param: logData: a dictionary of additional fields to include in the event
    :param: timestamp: the current timestamp
    :param: version: version of the event system to use ('1.0' is the only value at the time of writing)

    Keyword arguments for the AMQP connection. These do not need to be specified:

    :param: hostname: the hostname to connect to
    :param: port: the port to connect to
    :param: connection_attempts: the number of connection attempts to try
    :param: socket_timeout: the time to wait for a socket to connect
    :param: retry_delay: the time to wait between retrying the connection
    :param: username: the username to connect to as
    :param: password: the password to use when connecting

    :return: None
    """
    # first, connect
    _ensure_connected(**kwargs)

    # second, send the event
    _send(_format_message(message, **kwargs), **kwargs)


def _ensure_connected(**kwargs):
    global _connection, _channel
    if not _connection:
        config = CapoConfig(profile=kwargs.get('profile', None)).settings('edu.nrao.archive.configuration.AmqpServer')
        connection_parameters = pika.ConnectionParameters(
            host=kwargs.get('hostname', config.hostname),
            port=int(kwargs.get('port', config.port)),
            connection_attempts=kwargs.get('connection_attempts', 5),
            socket_timeout=kwargs.get('socket_timeout', 5000),
            retry_delay=kwargs.get('retry_delay', 500),
            credentials=
                pika.PlainCredentials(
                    username=kwargs.get('username', config.username),
                    password=kwargs.get('password', config.password)))
        _connection = pika.BlockingConnection(connection_parameters)
        _channel = _connection.channel()
        _channel.exchange_declare('archive.events',
                                      exchange_type='topic',
                                      durable=True, auto_delete=False)


def _format_message(message, **kwargs):
    default_message = {
        'application': 'python',
        'user': 'nobody',
        'request': None,
        'message': message,
        'logData': {},
        'timestamp': _datetime_to_dict(),
        'version': '1.0'
    }

    fields = ['application', 'user', 'request', 'message', 'logData',
              'timestamp', 'version']

    return {k: kwargs.get(k, default_message[k]) for k in fields}


def _send(message, **kwargs):
    _channel.basic_publish(exchange=EVENTS_EXCHANGE,
                           routing_key=kwargs.get('routing_key', '{}.event'.format(message['application'])),
                           body=json.dumps(message))


def _datetime_to_dict():
    # Turn 'now' in UTC into the date format events expect to speak.
    d, t = dict(), dict()
    now = datetime.datetime.utcnow()
    d['year'], d['month'], d['day'] = now.year, now.month, now.day
    t['hour'], t['minute'], t['second'], t['nano'] = now.hour, now.minute, now.second, now.microsecond * 1000
    return {'date': d, 'time': t}